﻿// Copyright Epic Games, Inc. All Rights Reserved.

#pragma once
#include <cstdint>
#include <type_traits>

inline uint32_t murmur_finalize32(uint32_t hash) {
    hash ^= hash >> 16;
    hash *= 0x85ebca6b;
    hash ^= hash >> 13;
    hash *= 0xc2b2ae35;
    hash ^= hash >> 16;
    return hash;
}

/**
 * Combines two hash values to get a third.
 * Note - this function is not commutative.
 *
 * This function cannot change for backward compatibility reasons.
 * You may want to choose HashCombineFast for a better in-memory hash combining function.
 */
[[nodiscard]] inline uint32_t hash_combine(uint32_t a, uint32_t c) {
    uint32_t b = 0x9e3779b9;
    a += b;

    a -= b; a -= c; a ^= (c>>13);
    b -= c; b -= a; b ^= (a<<8);
    c -= a; c -= b; c ^= (b>>13);
    a -= b; a -= c; a ^= (c>>12);
    b -= c; b -= a; b ^= (a<<16);
    c -= a; c -= b; c ^= (b>>5);
    a -= b; a -= c; a ^= (c>>3);
    b -= c; b -= a; b ^= (a<<10);
    c -= a; c -= b; c ^= (b>>15);

    return c;
}

/**
 * Combines two hash values to get a third.
 * Note - this function is not commutative.
 *
 * WARNING!  This function is subject to change and should only be used for creating
 *           combined hash values which don't leave the running process,
 *           e.g. GetTypeHash() overloads.
 */
[[nodiscard]] inline uint32_t hash_combine_fast(uint32_t a, uint32_t b) {
    return a ^ (b + 0x9e3779b9 + (a << 6) + (a >> 2));
}

[[nodiscard]] inline uint32_t pointer_hash(const void* key) {
    // Ignoring the lower 4 bits since they are likely zero anyway.
    // Higher bits are more significant in 64 bit builds.
    const uintptr_t ptr_int = reinterpret_cast<uintptr_t>(key) >> 4;
    return murmur_finalize32((uint32_t)ptr_int);
}

[[nodiscard]] inline uint32_t pointer_hash(const void* key, uint32_t c) {
    // we can use HashCombineFast here because pointers are non-persistent
    return hash_combine_fast(pointer_hash(key), c);
}

//
// Hash functions for common types.
//
// WARNING!  GetTypeHash result values are not expected to leave the running process.
//           Do not persist them to disk, send them to another running process or
//           expect them to be consistent across multiple runs.
//
template <
	typename ScalarType,
	std::enable_if_t<std::is_scalar_v<ScalarType> && !std::is_same_v<ScalarType, char32_t*> && !std::is_same_v<ScalarType, const char32_t*>>* = nullptr
>
[[nodiscard]] uint32_t get_type_hash(ScalarType value)
{
	if constexpr (std::is_integral_v<ScalarType>)
	{
		if constexpr (sizeof(ScalarType) <= 4)
		{
			return value;
		}
		else if constexpr (sizeof(ScalarType) == 8)
		{
			return (uint32_t)value + ((uint32_t)(value >> 32) * 23);
		}
		else if constexpr (sizeof(ScalarType) == 16)
		{
			const uint64_t low = (uint64_t)value;
			const uint64_t high = (uint64_t)(value >> 64);
			return get_type_hash(low) ^ get_type_hash(high);
		}
		else
		{
			static_assert(sizeof(ScalarType) == 0, "Unsupported integral type");
			return 0;
		}
	}
	else if constexpr (std::is_floating_point_v<ScalarType>)
	{
		if constexpr (std::is_same_v<ScalarType, float>)
		{
			return *(uint32_t*)&value;
		}
		else if constexpr (std::is_same_v<ScalarType, double>)
		{
			return get_type_hash(*(uint64_t*)&value);
		}
		else
		{
			static_assert(sizeof(ScalarType) == 0, "Unsupported floating point type");
			return 0;
		}
	}
	else if constexpr (std::is_enum_v<ScalarType>)
	{
		return get_type_hash((__underlying_type(ScalarType))value);
	}
	else if constexpr (std::is_pointer_v<ScalarType>)
	{
		// Once the TCHAR* deprecations below are removed, we want to prevent accidental string hashing, so this static_assert should be commented back in
		//static_assert(!TIsCharType<std::remove_pointer_t<ScalarType>>::Value, "Pointers to string types should use a PointerHash() or FCrc::Stricmp_DEPRECATED() call depending on requirements");

		return pointer_hash(value);
	}
	else
	{
		static_assert(sizeof(ScalarType) == 0, "Unsupported scalar type");
		return 0;
	}
}

template <
	typename T,
	uint32_t N,
	std::enable_if_t<!std::is_same_v<const T, const char32_t>>* = nullptr
>
uint32_t get_type_hash(T (&array)[N])
{
	return pointer_hash(array);
}

// template <
// 	typename T,
// 	std::enable_if_t<std::is_same_v<const T, const char32_t>>* = nullptr
// >
// uint32_t get_type_hash(T* value)
// {
// 	// Hashing a TCHAR* array differently from a void* is dangerous and is deprecated.
// 	// When removing these overloads post-deprecation, comment in the related static_assert in the std::is_pointer_v block of the GetTypeHash overload above.
// 	return FCrc::Strihash_DEPRECATED(value);
// }

template <typename T>
[[nodiscard]] uint32_t get_array_hash(const T* ptr, uint64_t size, uint32_t previous_hash = 0)
{
	uint32_t result = previous_hash;
	while (size)
	{
		result = hash_combine_fast(result, get_type_hash(*ptr));
		++ptr;
		--size;
	}

	return result;
}

// Use this when inside type that has get_type_hash() (no in-parameters) implemented. It makes GetTypeHash dispatch in global namespace
template <typename T>
[[nodiscard]] uint32_t get_type_hash_helper(const T& V) { return get_type_hash(V); }
