#include "device_selector.h"

void device_selector::classify_and_sort_devices(const vk::Instance instance, const vk::SurfaceKHR surface) {
	const auto& physical_devices = instance.enumeratePhysicalDevices();

	// 获取设备的评分列表
	auto device_scores = get_device_scores(physical_devices, surface);

	// 按评分排序设备（从高到低）
	auto sort_pred = [](const auto& a, const auto& b) { return a.second > b.second; };
	std::ranges::sort(device_scores, sort_pred);

	// 根据排序结果将设备分组
	group_devices(device_scores, surface);
}

vk::PhysicalDevice device_selector::get_main_device() const {
	// 优先使用同时支持显示和计算的设备
	if (!combined_group.empty()) {
		return combined_group[0];
	}
	// 如果没有同时支持显示和计算的设备，则使用显示设备
	if (!display_group.empty()) {
		return display_group[0];
	}
	// 如果没有显示设备，则使用计算设备
	if (!compute_group.empty()) {
		spdlog::warn("没有找到同时支持显示和计算的设备，将使用计算设备: {}", compute_group[0].getProperties().deviceName.data());
		return compute_group[0];
	}
	return nullptr;
}

vk::PhysicalDevice device_selector::get_main_compute_device() const {
	// 优先使用同时支持显示和计算的设备
	if (!combined_group.empty()) {
		return combined_group[0];
	}
	// 如果没有同时支持显示和计算的设备，则使用计算设备
	if (!compute_group.empty()) {
		return compute_group[0];
	}
	// 如果没有计算设备，则使用显示设备
	if (!display_group.empty()) {
		spdlog::warn("没有找到计算设备，将使用显示设备: {}", display_group[0].getProperties().deviceName.data());
		return display_group[0];
	}
	return nullptr;
}

std::vector<uint32_t> device_selector::get_queue_families(vk::PhysicalDevice in_physical_device,
                                                          const vk::QueueFlags& in_flags) {
	std::vector<uint32_t> result;
	const auto queue_families = in_physical_device.getQueueFamilyProperties();
	for (uint32_t i = 0; i < queue_families.size(); ++i) {
		if (queue_families[i].queueFlags & in_flags) {
			result.push_back(i);
		}
	}
	return result;
}

std::vector<device_selector::device_score> device_selector::get_device_scores(
	const std::vector<vk::PhysicalDevice>& devices, const vk::SurfaceKHR surface) {
	std::vector<device_score> deviceScores;

	for (const auto& dev: devices) {
		int score = score_device(dev, surface);
		deviceScores.emplace_back(dev, score);
	}
	return deviceScores;
}

int device_selector::score_device(const vk::PhysicalDevice& dev, const vk::SurfaceKHR surface) {
	int score = 0;

	// 获取设备属性和特性
	const auto& properties = dev.getProperties();
	const auto& features   = dev.getFeatures();

	// 首选独立显卡
	if (properties.deviceType == vk::PhysicalDeviceType::eDiscreteGpu) {
		score += 1000;
	} else if (properties.deviceType == vk::PhysicalDeviceType::eIntegratedGpu) {
		score += 500;
	}

	// 查询队列族属性
	const auto& queue_props = dev.getQueueFamilyProperties();
	bool has_graphics       = false;
	bool has_compute        = false;
	bool can_present        = false;

	for (uint32_t i = 0; i < queue_props.size(); ++i) {
		const auto& prop = queue_props[i];

		if (prop.queueFlags & vk::QueueFlagBits::eGraphics) {
			has_graphics = true;
			score += 100;
		}
		if (prop.queueFlags & vk::QueueFlagBits::eCompute) {
			has_compute = true;
			score += 100;
		}

		// 检查呈现能力
		const VkBool32 present_support = dev.getSurfaceSupportKHR(i, surface);
		if (present_support == VK_TRUE) {
			can_present = true;
			score += 100;
		}
	}

	// 如果同时支持图形、计算和呈现，则加额外分
	if (has_graphics && has_compute && can_present) {
		score += 500; // 额外奖励
	}

	// 考虑内存大小
	const auto& memory_properties = dev.getMemoryProperties();
	uint64_t total_memory         = 0;
	for (const auto& heap : memory_properties.memoryHeaps) {
		if (heap.flags & vk::MemoryHeapFlagBits::eDeviceLocal) {
			total_memory += heap.size;
		}
	}
	score += static_cast<int>(total_memory / (1024 * 1024 * 1024)); // 每GB加1分

	std::string device_name = properties.deviceName;
	spdlog::info("{}: {}分", device_name, score);
	return score;
}

void device_selector::group_devices(const std::vector<device_score>& deviceScores, const vk::SurfaceKHR surface) {
	for (const auto& [dev, score] : deviceScores) {
		const device_type& type = classify_device(dev, surface);

		switch (type) {
			case device_type::combined:
				combined_group.push_back(dev);
				display_group.push_back(dev);
				compute_group.push_back(dev);
				break;
			case device_type::display:
				display_group.push_back(dev);
				break;
			case device_type::compute:
				compute_group.push_back(dev);
				break;
			default:
				// 忽略不符合条件的设备
				break;
		}
	}
}

device_selector::device_type device_selector::classify_device(const vk::PhysicalDevice& dev,
	const vk::SurfaceKHR surface) {
	bool has_graphics = false;
	bool has_compute  = false;
	bool can_present  = false;

	const auto& queue_props = dev.getQueueFamilyProperties();

	for (uint32_t i = 0; i < queue_props.size(); ++i) {
		const auto& prop = queue_props[i];

		if (prop.queueFlags & vk::QueueFlagBits::eGraphics) {
			has_graphics = true;
		}
		if (prop.queueFlags & vk::QueueFlagBits::eCompute) {
			has_compute = true;
		}

		// 检查呈现能力
		const VkBool32 present_support = dev.getSurfaceSupportKHR(i, surface);
		if (present_support == VK_TRUE) {
			can_present = true;
		}
	}

	if (has_graphics && has_compute && can_present) {
		return device_type::combined;
	}
	if (has_graphics && can_present) {
		return device_type::display;
	}
	if (has_compute) {
		return device_type::compute;
	}
	return device_type::unknown;
}
