#include "renderer.h"

#include <spdlog/spdlog.h>

#include "window/renderer_window.h"
#define STB_IMAGE_IMPLEMENTATION
#define STBI_FAILURE_USERMSG
#include "aorii.h"
#include "renderer_shader.h"
#include "renderer_text.h"
#include "async/thread_pool.h"
#include "window/window_manager.h"
#include "misc/stb_image.h"

aorii_renderer* s_renderer{};

// renderer_texture* renderer::load_image(const std::string& file_path, texture_format in_format) {
//     int width, height, channels;
//     const auto result = stbi_load(file_path.c_str(), &width, &height, &channels, 0);
//     if (!result) {
//         const char* reason = stbi_failure_reason();
//         spdlog::critical("加载图片 {0} 失败: {1}", file_path, reason);
//         return nullptr;
//     }
//     int target_channel_count = get_format_channel_count(in_format);
//     unsigned int row_pitch = 0;
//
//     auto texture = create_texture(width, height, in_format, device_memory_type::GPU);
//     auto data = texture->lock(&row_pitch);
//     for (int y = 0; y < height; ++y) {
//         for (int x = 0; x < width; ++x) {
//             unsigned char* src = result + (y * width + x) * channels;
//             unsigned char* dst = static_cast<unsigned char*>(data) + y * row_pitch + x * target_channel_count;
//             switch (channels) {
//             case 1:
//                 dst[0] = src[0];
//                 dst[1] = 0;
//                 dst[2] = 0;
//                 dst[3] = 255;
//                 break;
//             case 2:
//                 dst[0] = src[0];
//                 dst[1] = src[1];
//                 dst[2] = 0;
//                 dst[3] = 255;
//                 break;
//             case 3:
//                 dst[0] = src[0];
//                 dst[1] = src[1];
//                 dst[2] = src[2];
//                 dst[3] = 255;
//                 break;
//             case 4:
//                 dst[0] = src[0];
//                 dst[1] = src[1];
//                 dst[2] = src[2];
//                 dst[3] = src[3];
//                 break;
//             default:
//                 break;
//             }
//         }
//     }
//     texture->unlock();
//     stbi_image_free(result);
//
//     return texture;
// }

aorii_renderer* aorii::get_renderer() {
    return s_renderer;
}

bool aorii::create_renderer() {
    if (s_renderer) return true;

    s_renderer = new aorii_renderer();
    if (!s_renderer->init()) {
        delete s_renderer;
        s_renderer = nullptr;
        spdlog::error("创建渲染器失败");
        return false;
    }
    return s_renderer != nullptr;
}

void aorii::destroy_renderer() {
    if (!s_renderer) return;

    s_renderer->destroy();
    delete s_renderer;
}

bool aorii_renderer::init() {
    spdlog::info("初始化渲染器");
    uint32_t vulkan_api_version = 0;

    if (vk::enumerateInstanceVersion(&vulkan_api_version) != vk::Result::eSuccess) {
        spdlog::warn("无法获取 Vulkan 版本号, 将使用默认版本 1.0.0");
        vulkan_api_version = VK_API_VERSION_1_0;
    }

    vk::ApplicationInfo app_info{};
    app_info.setPEngineName("Aorii");
    app_info.setEngineVersion(VK_MAKE_VERSION(AORII_VERSION_MAJOR, AORII_VERSION_MINOR, AORII_VERSION_PATCH));
    app_info.setApiVersion(vulkan_api_version);
    app_info.setPApplicationName("AoriiHelloWorld");
    app_info.setApplicationVersion(VK_MAKE_VERSION(1, 0, 0));
    spdlog::info("Vulkan 版本: {}.{}.{}", VK_VERSION_MAJOR(app_info.apiVersion), VK_VERSION_MINOR(app_info.apiVersion), VK_VERSION_PATCH(app_info.apiVersion));

    vk::InstanceCreateInfo create_info{};
    create_info.setPApplicationInfo(&app_info);

    std::vector<const char*> extensions;
    {
#if DEBUG
        extensions.push_back(VK_EXT_DEBUG_REPORT_EXTENSION_NAME);
        extensions.push_back(VK_EXT_DEBUG_UTILS_EXTENSION_NAME);
#endif
        // 交换链
        uint32_t glfw_extension_count = 0;
        auto glfw_extensions = glfwGetRequiredInstanceExtensions(&glfw_extension_count);
        for (int i = 0; i < glfw_extension_count; ++i) {
            extensions.push_back(glfw_extensions[i]);
        }
        for (auto extension: extensions) {
            spdlog::info("启用扩展: {}", extension);
        }

        create_info.setPEnabledExtensionNames(extensions);
    }
    std::vector<const char*> layers;
    {
#if DEBUG
        layers.push_back("VK_LAYER_KHRONOS_validation");
        create_info.setPEnabledLayerNames(layers);
        for (auto layer: layers) {
            spdlog::info("启用层: {}", layer);
        }
#endif
    }

    instance = createInstance(create_info);
    if (!instance) {
        spdlog::critical("创建 Vulkan 实例失败");
        return false;
    }

    const auto& main_window = aorii::get_window_manager()->get_main_window().lock();
    if (!main_window->create_surface())
        return false;
    selector.classify_and_sort_devices(instance, main_window->get_surface());

    std::vector<const char*> device_extensions;
    device_extensions.push_back(VK_KHR_SWAPCHAIN_EXTENSION_NAME);

    std::vector<vk::DeviceQueueCreateInfo> queue_create_infos;
    const auto family_indexes = device_selector::get_queue_families(selector.get_main_device(), vk::QueueFlagBits::eGraphics);
    for (const auto i : family_indexes) {
        constexpr float queue_priority = 1.0f;
        vk::DeviceQueueCreateInfo queue_create_info{};
        queue_create_info.setQueueFamilyIndex(i);
        queue_create_info.setQueueCount(1);
        queue_create_info.setPQueuePriorities(&queue_priority);
        queue_create_infos.push_back(queue_create_info);
        break;
    }

    vk::DeviceCreateInfo device_create_info{};
    device_create_info.setPEnabledExtensionNames(device_extensions);
    device_create_info.setQueueCreateInfos(queue_create_infos);

    main_device.create_device(selector.get_main_device(), device_create_info);
    spdlog::info("Vulkan 主设备使用: {}", main_device.get_physical_device_name());
    for (auto extension: device_extensions) {
        spdlog::info("启用设备扩展: {}", extension);
    }

    renderer_swapchain::create_info swapchain_info{};
    if (!main_window->create_swap_chain(swapchain_info))
        return false;

    renderer_shader frag_shader;
    frag_shader.load("resource/shaders/test.frag.spv", main_device);
    if (!frag_shader.is_valid()) {
        return false;
    }
    renderer_shader vert_shader;
    vert_shader.load("resource/shaders/test.vert.spv", main_device);
    if (!vert_shader.is_valid()) {
        return false;
    }

    return true;
}

void aorii_renderer::destroy() {
    spdlog::info("销毁渲染器");
    main_device->destroy();
    instance.destroy();
}

vk::SurfaceKHR aorii_renderer::create_surface(GLFWwindow* in_window) const {
    VkSurfaceKHR surface;
    if (instance && glfwCreateWindowSurface(instance, in_window, nullptr, &surface) != VK_SUCCESS) {
        spdlog::critical("创建窗口表面失败");
        return nullptr;
    }
    return { surface };
}

void aorii_renderer::destroy_surface(const vk::SurfaceKHR& in_surface) const {
    instance.destroySurfaceKHR(in_surface);
}

const std::vector<device_handle>& aorii_renderer::get_compute_device() const {
    if (!compute_device.empty()) return compute_device;

    for (const auto& device: selector.compute_group) {
        constexpr float queue_priority = 1.0f;
        std::vector<vk::DeviceQueueCreateInfo> queue_create_infos;
        const auto& families = device_selector::get_queue_families(device, vk::QueueFlagBits::eCompute);
        for (const auto i: families) {
            vk::DeviceQueueCreateInfo queue_create_info{};
            queue_create_info.setQueueFamilyIndex(i);
            queue_create_info.setQueueCount(1);
            queue_create_info.setPQueuePriorities(&queue_priority);
            queue_create_infos.push_back(queue_create_info);
        }

        device_handle handle;
        vk::DeviceCreateInfo device_create_info{};
        device_create_info.setQueueCreateInfos(queue_create_infos);
        handle.create_device(device, device_create_info);
        compute_device.push_back(handle);
    }
    return compute_device;
}

