#pragma once
#include <span>
#include <spdlog/spdlog.h>

#include "renderer_types.h"

class buffer_memory_strategy;

class renderer_buffer {
public:
    explicit renderer_buffer(const device_handle& in_device, const renderer_buffer_desc& in_desc);

    virtual ~renderer_buffer();

    [[nodiscard]] const auto& get_memory() const {
        return memory;
    }
    [[nodiscard]] const auto& get_desc() const {
        return desc;
    }
    [[nodiscard]] auto get_size() const {
        return desc.size;
    }
    [[nodiscard]] auto get_buffer() const {
        return buffer;
    }

    template<class T>
    void set_data(const T& in_data, vk::CommandBuffer in_command_buffer = nullptr) {
        copy_from(&in_data, sizeof(T), in_command_buffer);
    }
    template<class T>
    void set_data(const std::span<const T>& in_data, vk::CommandBuffer in_command_buffer = nullptr) {
        copy_from(in_data.data(), in_data.size_bytes(), in_command_buffer);
    }
    void copy_from(const void* in_data, vk::DeviceSize in_size, vk::CommandBuffer in_command_buffer = nullptr);

    /**
     * @brief 重新分配缓冲区大小
     * @param in_new_size 新的大小
     * @param in_command_buffer 命令缓冲区
     * @param in_preserve_data 是否保留原有数据
     * @param in_allow_in_place 是否允许原地扩展
     * @param in_growth_factor 扩展因子
     */
    void resize(vk::DeviceSize in_new_size, vk::CommandBuffer in_command_buffer, bool in_preserve_data = true, bool in_allow_in_place = true, float in_growth_factor = 1.5f);
protected:
    device_handle device;
    renderer_buffer_desc desc;
    vk::Buffer buffer;
    vk::DeviceMemory memory;
    std::unique_ptr<buffer_memory_strategy> memory_strategy;
private:
    [[nodiscard]] auto find_memory_type(const uint32_t type_filter, const vk::MemoryPropertyFlags properties) const {
        const auto& mem_properties = device.physical_device.getMemoryProperties();

        for (uint32_t i = 0; i < mem_properties.memoryTypeCount; i++) {
            if ((type_filter & (1 << i)) && (mem_properties.memoryTypes[i].propertyFlags & properties) == properties) {
                return i;
            }
        }

        throw std::runtime_error("failed to find suitable memory type!");
    }

    vk::DeviceMemory allocate_memory(const vk::MemoryRequirements& in_requirements, vk::MemoryPropertyFlags props) {
        vk::MemoryAllocateInfo alloc_info;
        alloc_info.allocationSize  = in_requirements.size;
        alloc_info.memoryTypeIndex = find_memory_type(in_requirements.memoryTypeBits, props);

        return device->allocateMemory(alloc_info);
    }

    bool try_resize_in_place(vk::DeviceSize in_size) {
        const auto& mem_props = device->getBufferMemoryRequirements(buffer);
        return in_size <= mem_props.size;
    }
    [[nodiscard]] vk::DeviceSize calculate_new_size(const vk::DeviceSize in_new_size, const float in_growth_factor) const {
        const vk::DeviceSize growth_size = get_size() * in_growth_factor;
        return std::max(in_new_size, growth_size);
    }
};
