#include "renderer_buffer_memory_strategy.h"

#include "renderer_buffer.h"

std::unique_ptr<buffer_memory_strategy> buffer_memory_strategy::create(vk::MemoryPropertyFlags in_memory_props) {
	if (in_memory_props & vk::MemoryPropertyFlagBits::eHostVisible) {
		return std::make_unique<host_visible_strategy>();
	}
	if (in_memory_props & vk::MemoryPropertyFlagBits::eDeviceLocal) {
		return std::make_unique<device_local_strategy>();
	}
	throw std::runtime_error("Unsupported memory property flags");
}

void host_visible_strategy::copy_from(const device_handle& in_device, vk::Buffer in_buffer, const vk::DeviceMemory in_device_memory,
	const void* in_data, const vk::DeviceSize in_size, vk::CommandBuffer in_command_buffer) {
	void* mapped = in_device->mapMemory(in_device_memory, 0, in_size);
	std::memcpy(mapped, in_data, in_size);

	vk::MappedMemoryRange range;
	range.memory = in_device_memory;
	range.offset = 0;
	range.size   = in_size;
	in_device->flushMappedMemoryRanges(range);

	in_device->unmapMemory(in_device_memory);
}

void device_local_strategy::copy_from(const device_handle& in_device, vk::Buffer in_buffer, vk::DeviceMemory in_device_memory,
	const void* in_data, const vk::DeviceSize in_size, vk::CommandBuffer in_command_buffer) {

	// 创建暂存缓冲区
	const renderer_buffer staging_buffer(in_device, renderer_buffer_desc::staging_buffer(in_size));

	// 复制数据到暂存缓冲区
	void* mapped = in_device->mapMemory(staging_buffer.get_memory(), 0, in_size);
	std::memcpy(mapped, in_data, in_size);
	in_device->unmapMemory(staging_buffer.get_memory());

	// 记录复制命令
	vk::BufferCopy copy_region{};
	copy_region.size = in_size;
	in_command_buffer.copyBuffer(staging_buffer.get_buffer(), in_buffer, copy_region);
}
