#include "renderer_swapchain.h"

// std::string to_string(const vk::Extent2D& in) {
// 	return fmt::format("Extent2D: {}x{}", in.width, in.height);
// }

// std::string to_string(const vk::SwapchainCreateInfoKHR& in) {
// 	// 格式化字符串
// 	std::string result = fmt::format("SwapchainCreateInfoKHR:\n"
// 	                                 "  surface: {}\n"
// 	                                 "  minImageCount: {}\n"
// 	                                 "  imageFormat: {}\n"
// 	                                 "  imageColorSpace: {}\n"
// 	                                 "  imageExtent: {}\n"
// 	                                 "  imageArrayLayers: {}\n"
// 	                                 "  imageUsage: {}\n"
// 	                                 "  presentMode: {}\n"
// 	                                 "  preTransform: {}\n"
// 	                                 "  compositeAlpha: {}\n"
// 	                                 "  clipped: {}\n",
// 	                                 in.surface,
// 	                                 in.minImageCount,
// 	                                 vk::to_string(in.imageFormat),
// 	                                 vk::to_string(in.imageColorSpace),
// 	                                 to_string(in.imageExtent),
// 	                                 in.imageArrayLayers,
// 	                                 vk::to_string(in.imageUsage),
// 	                                 vk::to_string(in.presentMode),
// 	                                 vk::to_string(in.preTransform),
// 	                                 vk::to_string(in.compositeAlpha),
// 	                                 in.clipped);
// 	return result;
// }

bool renderer_swapchain::init_swapchain(const device_handle& in_device, const vk::SurfaceKHR& in_surface, const create_info& in_info) {
	device  = in_device;
	surface = in_surface;
	info    = in_info;
	// 获取交换链支持的信息
	const auto& surface_capabilities = device.physical_device.getSurfaceCapabilitiesKHR(in_surface);

	// 验证并调整图像数量
	uint32_t image_count = std::max(in_info.min_image_count,
	                                surface_capabilities.minImageCount);
	if (surface_capabilities.maxImageCount > 0) {
		image_count = std::clamp(image_count,
		                         surface_capabilities.minImageCount,
		                         surface_capabilities.maxImageCount);
	}

	// 验证并调整图像格式
	auto surface_formats = device.physical_device.getSurfaceFormatsKHR(in_surface);
	if (surface_formats.empty()) {
		spdlog::error("获取表面格式失败");
		return false;
	}
	if (surface_formats.size() == 1 && surface_formats[0].format == vk::Format::eUndefined) {
		info.format = vk::Format::eB8G8R8A8Srgb;
		info.color_space = vk::ColorSpaceKHR::eSrgbNonlinear;
		spdlog::warn("表面格式未指定，使用默认格式 {} {}", vk::to_string(info.format), vk::to_string(info.color_space));
	} else {
		bool found = false;
		for (const auto& format : surface_formats) {
			if (format.format == info.format && format.colorSpace == info.color_space) {
				found = true;
				break;
			}
		}
		if (!found) {
			info.format = surface_formats[0].format;
			info.color_space = surface_formats[0].colorSpace;
			spdlog::warn("请求的表面格式不支持，使用默认格式 {} {}", vk::to_string(info.format), vk::to_string(info.color_space));
		}
	}

	// 验证并调整呈现模式
	auto present_modes = device.physical_device.getSurfacePresentModesKHR(in_surface);
	if (present_modes.empty()) {
		spdlog::error("获取呈现模式失败");
		return false;
	}
	if (std::ranges::find(present_modes, info.present_mode) == present_modes.end()) {
		info.present_mode = vk::PresentModeKHR::eFifo;
		spdlog::warn("请求的呈现模式不支持，使用默认呈现模式 Fifo");
	}

	// 验证并调整交换链尺寸
	info.extent = std::clamp(info.extent,
	                         surface_capabilities.minImageExtent,
	                         surface_capabilities.maxImageExtent);

	// 验证并调整图像用途
	if ((surface_capabilities.supportedUsageFlags & info.image_usage) != info.image_usage) {
		info.image_usage = vk::ImageUsageFlagBits::eColorAttachment;
		spdlog::warn("请求的图像用途不支持，使用默认图像用途 ColorAttachment");
	}

	// 验证并调整合成 alpha
	if (!(surface_capabilities.supportedCompositeAlpha & vk::CompositeAlphaFlagBitsKHR::eOpaque)) {
		spdlog::warn("不支持不透明合成 alpha");
		return false;
	}

	// 创建交换链信息
	vk::SwapchainCreateInfoKHR create_info{};
	create_info.surface          = surface;
	create_info.minImageCount    = image_count;
	create_info.imageFormat      = info.format;
	create_info.imageColorSpace  = info.color_space;
	create_info.imageExtent      = info.extent;
	create_info.imageArrayLayers = 1;
	create_info.imageUsage       = info.image_usage;
	create_info.imageSharingMode = vk::SharingMode::eExclusive;
	create_info.presentMode      = info.present_mode;
	create_info.preTransform     = surface_capabilities.currentTransform;
	create_info.compositeAlpha   = vk::CompositeAlphaFlagBitsKHR::eOpaque;
	create_info.clipped          = VK_TRUE;
	create_info.imageExtent      = info.extent;

	// 创建交换链
	swapchain = device->createSwapchainKHR(create_info);
	if (!swapchain) {
		spdlog::error("创建交换链失败");
		return false;
	}
	images = device->getSwapchainImagesKHR(swapchain);
	if (images.empty()) {
		spdlog::error("获取交换链图像失败");
		return false;
	}

	// 创建图像视图
	return create_image_views(device.device, in_info.format);
}

bool renderer_swapchain::create_image_views(const vk::Device& device, vk::Format format) {
	image_views.resize(images.size());

	for (size_t i = 0; i < images.size(); i++) {
		vk::ImageViewCreateInfo view_info{};
		view_info.image                           = images[i];
		view_info.viewType                        = vk::ImageViewType::e2D;
		view_info.format                          = format;
		view_info.components.r                    = vk::ComponentSwizzle::eIdentity;
		view_info.components.g                    = vk::ComponentSwizzle::eIdentity;
		view_info.components.b                    = vk::ComponentSwizzle::eIdentity;
		view_info.components.a                    = vk::ComponentSwizzle::eIdentity;
		view_info.subresourceRange.aspectMask     = vk::ImageAspectFlagBits::eColor;
		view_info.subresourceRange.baseMipLevel   = 0;
		view_info.subresourceRange.levelCount     = 1;
		view_info.subresourceRange.baseArrayLayer = 0;
		view_info.subresourceRange.layerCount     = 1;

		auto new_view = device.createImageView(view_info);
		if (new_view)
			image_views[i] = new_view;
		else
			goto ERROR;
	}
	return true;

ERROR:
	spdlog::error("创建图像视图失败");
	for (const auto view: image_views) { device.destroyImageView(view); }
	return false;
}

void renderer_swapchain::cleanup() {
	if (device) {
		for (const auto& view : image_views) {
			device->destroyImageView(view);
		}
		device->destroySwapchainKHR(swapchain);
	}
}
