#pragma once
#include <vulkan/vulkan.hpp>
#include <spdlog/spdlog.h>

struct device_handle {
    vk::PhysicalDevice physical_device;
    vk::Device device;

    // 重载->运算符，使得device_set可以直接调用vk::Device的成员函数
    vk::Device* operator->() { return &device; }
    const vk::Device* operator->() const { return &device; }

    operator vk::PhysicalDevice() const { return physical_device; }
    operator vk::Device() const { return device; }
    operator bool() const { return static_cast<bool>(device); }
    bool operator!() const { return !device; }

    void create_device(const vk::PhysicalDevice in_physical_device, const vk::DeviceCreateInfo& in_create_info) {
        physical_device = in_physical_device;
        device = physical_device.createDevice(in_create_info);
        spdlog::info("创建 Vulkan 设备: {}", get_physical_device_name());
    }
    void destroy_device() {
        if (device) {
            device.destroy();
            device = nullptr;
        }
    }

    // 获取物理设备GUID
    [[nodiscard]] uint32_t get_physical_device_id() const;
    // 获取物理设备名称
    [[nodiscard]] std::string get_physical_device_name() const;
};

struct renderer_buffer_desc {
    // 大小
    vk::DeviceSize size;
    // 缓冲使用标志
    vk::BufferUsageFlags usage_flags;
    // 内存属性
    vk::MemoryPropertyFlags memory_props;
    // 内存共享标志
    bool concurrent = false;
    // 队列族
    std::vector<uint32_t> queue_families;

    [[nodiscard]] bool is_host_visible() const {
        return (memory_props & vk::MemoryPropertyFlagBits::eHostVisible)
                == vk::MemoryPropertyFlagBits::eHostVisible;
    }

    [[nodiscard]] bool is_device_local() const {
        return (memory_props & vk::MemoryPropertyFlagBits::eDeviceLocal)
                == vk::MemoryPropertyFlagBits::eDeviceLocal;
    }

    operator vk::BufferCreateInfo() const;
    operator vk::BufferCreateInfo() {
        return static_cast<const renderer_buffer_desc&>(*this);
    }

    static renderer_buffer_desc staging_buffer(const vk::DeviceSize in_size) {
        return {
            .size         = in_size,
            .usage_flags  = vk::BufferUsageFlagBits::eTransferSrc,
            .memory_props = vk::MemoryPropertyFlagBits::eHostVisible | vk::MemoryPropertyFlagBits::eHostCoherent
        };
    }

    static renderer_buffer_desc uniform_buffer(const vk::DeviceSize in_size) {
        return {
            .size         = in_size,
            .usage_flags  = vk::BufferUsageFlagBits::eUniformBuffer,
            .memory_props = vk::MemoryPropertyFlagBits::eHostVisible | vk::MemoryPropertyFlagBits::eHostCoherent
        };
    }

    static renderer_buffer_desc index_buffer(const vk::DeviceSize in_size) {
        return {
            .size         = in_size,
            .usage_flags  = vk::BufferUsageFlagBits::eIndexBuffer | vk::BufferUsageFlagBits::eTransferDst,
            .memory_props = vk::MemoryPropertyFlagBits::eDeviceLocal
        };
    }

    static renderer_buffer_desc vertex_buffer(const vk::DeviceSize in_size) {
        return {
            .size         = in_size,
            .usage_flags  = vk::BufferUsageFlagBits::eVertexBuffer | vk::BufferUsageFlagBits::eTransferDst,
            .memory_props = vk::MemoryPropertyFlagBits::eDeviceLocal
        };
    }
};

