#include "renderer_window.h"

#include <spdlog/spdlog.h>

#include "window_manager.h"
#include <GLFW/glfw3native.h>

renderer_window::renderer_window(const window_desc& in_desc) {
	glfwWindowHint(GLFW_RESIZABLE, in_desc.resizable);
	glfwWindowHint(GLFW_DECORATED, in_desc.decorated);
	glfwWindowHint(GLFW_VISIBLE, in_desc.visible);
	glfwWindowHint(GLFW_FOCUSED, in_desc.focused);
	glfwWindowHint(GLFW_FOCUS_ON_SHOW, in_desc.focus_on_show);
	glfwWindowHint(GLFW_MAXIMIZED, in_desc.maximized);
	glfwWindowHint(GLFW_FLOATING, in_desc.floating);
	glfwWindowHint(GLFW_CENTER_CURSOR, in_desc.center_cursor);
	glfwWindowHint(GLFW_TRANSPARENT_FRAMEBUFFER, in_desc.transparent);

	window = glfwCreateWindow(in_desc.resolution.width, in_desc.resolution.height, in_desc.title.c_str(), in_desc.monitor, nullptr);
	glfwSetWindowUserPointer(window, this);
	if (!window) {
		spdlog::error("创建 GLFW 窗口失败");
		return;
	}

	glfwSetWindowSizeCallback(window, [](GLFWwindow* in_window, int in_width, int in_height) {
		const auto my_window = static_cast<renderer_window*>(glfwGetWindowUserPointer(in_window));
		my_window->on_resize(in_width, in_height);
	});
	glfwSetWindowCloseCallback(window, [](GLFWwindow* in_window) {
		const auto my_window = static_cast<renderer_window*>(glfwGetWindowUserPointer(in_window));
		my_window->on_glfw_close();
	});

	if (in_desc.create_surface) {
		create_surface();
	}
}

renderer_window::~renderer_window() {
	delete swapchain;
	if (surface) {
		const auto renderer = aorii::get_renderer();
		if (renderer) {
			renderer->destroy_surface(surface);
		}
	}
}

void renderer_window::set_title(const std::string& in_title) const {
	glfwSetWindowTitle(window, in_title.c_str());
}

void renderer_window::set_size(int in_width, int in_height) {
	glfwSetWindowSize(window, in_width, in_height);
	on_resize(in_width, in_height);
}

void renderer_window::close() const {
	glfwSetWindowShouldClose(window, GLFW_TRUE);
}

void renderer_window::update() {
	// swapchain->;
}

bool renderer_window::create_surface() {
	const auto renderer = aorii::get_renderer();
	if (!renderer)
		return false;
	surface = renderer->create_surface(window);
	if (!surface) {
		spdlog::error("创建窗口表面失败");
		return false;
	}
	return true;
}

bool renderer_window::create_swap_chain(const renderer_swapchain::create_info& in_create_info) {
	const auto renderer = aorii::get_renderer();
	if (!renderer)
		return false;
	if (!surface) {
		spdlog::error("窗口表面未创建");
		return false;
	}

	const auto& main_device = renderer->get_main_device();
	auto temp_swapchain_info = in_create_info;
	temp_swapchain_info.extent = get_size();

	swapchain = new renderer_swapchain();
	if (!swapchain->init_swapchain(main_device, surface, temp_swapchain_info)) {
		spdlog::error("创建交换链失败");
		return false;
	}
	return true;
}

void* renderer_window::get_native_handle() const {
#if defined(GLFW_EXPOSE_NATIVE_WIN32)
	return glfwGetWin32Window(window);
#elif defined(GLFW_EXPOSE_NATIVE_COCOA)
	return glfwGetCocoaWindow(window);
#elif defined(GLFW_EXPOSE_NATIVE_X11)
	return reinterpret_cast<void*>(
		static_cast<uintptr_t>(glfwGetX11Window(window)));
#elif defined(GLFW_EXPOSE_NATIVE_WAYLAND)
	return glfwGetWaylandWindow(window);
#elif defined(GLFW_EXPOSE_NATIVE_EGL)
	return static_cast<void*>(glfwGetEGLDisplay(window));
#else
	static_assert(false, "No native platform window implementation available");
	return nullptr;
#endif
}

void renderer_window::on_resize(int in_width, int in_height) {
	spdlog::info("窗口大小变更：{}x{}", in_width, in_height);
	if (swapchain) {
		swapchain->resize({static_cast<uint32_t>(in_width), static_cast<uint32_t>(in_height)});
	}
}

void renderer_window::on_glfw_close() {
	spdlog::info("窗口关闭");
	aorii::get_window_manager()->destroy_window(this);
}
